from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.consts import inline_query_result_types
from telebox.telegram.types.types.inline_keyboard_markup import InlineKeyboardMarkup
from telebox.telegram.types.types.input_message_content import InputMessageContent


@dataclass(unsafe_hash=True)
class InlineQueryResultLocation(Type):
    id: str
    latitude: float
    longitude: float
    title: str
    horizontal_accuracy: Optional[float] = None
    live_period: Optional[int] = None
    heading: Optional[int] = None
    proximity_alert_radius: Optional[int] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional[InputMessageContent] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None
    type: str = inline_query_result_types.LOCATION



from typing import Union

from telebox.telegram.types.types.bot_command_scope_default import (
    BotCommandScopeDefault
)
from telebox.telegram.types.types.bot_command_scope_all_private_chats import (
    BotCommandScopeAllPrivateChats
)
from telebox.telegram.types.types.bot_command_scope_all_group_chats import (
    BotCommandScopeAllGroupChats
)
from telebox.telegram.types.types.bot_command_scope_all_chat_administrators import (
    BotCommandScopeAllChatAdministrators
)
from telebox.telegram.types.types.bot_command_scope_chat import (
    BotCommandScopeChat
)
from telebox.telegram.types.types.bot_command_scope_chat_administrators import (
    BotCommandScopeChatAdministrators
)
from telebox.telegram.types.types.bot_command_scope_chat_member import (
    BotCommandScopeChatMember
)


BotCommandScope = Union[BotCommandScopeDefault,
                        BotCommandScopeAllPrivateChats,
                        BotCommandScopeAllGroupChats,
                        BotCommandScopeAllChatAdministrators,
                        BotCommandScopeChat,
                        BotCommandScopeChatAdministrators,
                        BotCommandScopeChatMember]



from typing import Union, Optional, Any, Literal, TypeVar, Type
from datetime import timezone, datetime as datetime_
from dataclasses import is_dataclass
import secrets

from requests import Session, Response
from dataclass_factory import Factory, Schema

from telebox.telegram.errors import get_error
from telebox.telegram.consts import chat_member_statuses
from telebox.telegram.types.types.response_parameters import ResponseParameters
from telebox.telegram.types.types.update import Update
from telebox.telegram.types.types.webhook_info import WebhookInfo
from telebox.telegram.types.types.user import User
from telebox.telegram.types.types.message import Message
from telebox.telegram.types.types.message_entity import MessageEntity
from telebox.telegram.types.types.inline_keyboard_markup import InlineKeyboardMarkup
from telebox.telegram.types.types.reply_keyboard_markup import ReplyKeyboardMarkup
from telebox.telegram.types.types.reply_keyboard_remove import ReplyKeyboardRemove
from telebox.telegram.types.types.force_reply import ForceReply
from telebox.telegram.types.types.input_file import InputFile
from telebox.telegram.types.types.message_id import MessageId
from telebox.telegram.types.types.input_media import InputMedia
from telebox.telegram.types.types.input_media_audio import InputMediaAudio
from telebox.telegram.types.types.input_media_document import InputMediaDocument
from telebox.telegram.types.types.input_media_photo import InputMediaPhoto
from telebox.telegram.types.types.input_media_video import InputMediaVideo
from telebox.telegram.types.types.user_profile_photos import UserProfilePhotos
from telebox.telegram.types.types.chat_permissions import ChatPermissions
from telebox.telegram.types.types.file import File
from telebox.telegram.types.types.chat_invite_link import ChatInviteLink
from telebox.telegram.types.types.chat import Chat
from telebox.telegram.types.types.bot_command import BotCommand
from telebox.telegram.types.types.bot_command_scope import BotCommandScope
from telebox.telegram.types.types.menu_button import MenuButton
from telebox.telegram.types.types.chat_administrator_rights import ChatAdministratorRights
from telebox.telegram.types.types.poll import Poll
from telebox.telegram.types.types.sticker import Sticker
from telebox.telegram.types.types.sticker_set import StickerSet
from telebox.telegram.types.types.mask_position import MaskPosition
from telebox.telegram.types.types.inline_query_result import InlineQueryResult
from telebox.telegram.types.types.sent_web_app_message import SentWebAppMessage
from telebox.telegram.types.types.labeled_price import LabeledPrice
from telebox.telegram.types.types.shipping_option import ShippingOption
from telebox.telegram.types.types.passport_element_error import PassportElementError
from telebox.telegram.types.types.game_high_score import GameHighScore
from telebox.telegram.types.types.chat_member import ChatMember
from telebox.telegram.types.types.chat_member_owner import ChatMemberOwner
from telebox.telegram.types.types.chat_member_administrator import ChatMemberAdministrator
from telebox.telegram.types.types.chat_member_member import ChatMemberMember
from telebox.telegram.types.types.chat_member_restricted import ChatMemberRestricted
from telebox.telegram.types.types.chat_member_left import ChatMemberLeft
from telebox.telegram.types.types.chat_member_banned import ChatMemberBanned


URL = "https://api.telegram.org"
DataclassObject = TypeVar("DataclassObject")
_CHAT_MEMBER_TYPES = {
    chat_member_statuses.CREATOR: ChatMemberOwner,
    chat_member_statuses.ADMINISTRATOR: ChatMemberAdministrator,
    chat_member_statuses.MEMBER: ChatMemberMember,
    chat_member_statuses.RESTRICTED: ChatMemberRestricted,
    chat_member_statuses.LEFT: ChatMemberLeft,
    chat_member_statuses.KICKED: ChatMemberBanned
}


def _convert_datetime_to_timestamp(datetime: Optional[datetime_]) -> int:
    if datetime is not None:
        return int(datetime.timestamp())
    else:
        return 0


def _convert_timestamp_to_datetime(timestamp: int) -> Optional[datetime_]:
    if timestamp:
        return datetime_.fromtimestamp(timestamp, tz=timezone.utc)


_dataclass_factory = Factory(
    schemas={
        datetime_: Schema(
            serializer=_convert_datetime_to_timestamp,
            parser=_convert_timestamp_to_datetime
        )
    }
)


class Telegram:

    def __init__(self, session: Session, token: str, *, url: str = URL, is_local: bool = False):
        self._session = session
        self._token = token
        self._url = url
        self._is_local = is_local

    def get_updates(
        self,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        timeout: Optional[int] = None,
        allowed_updates: Optional[list[str]] = None
    ) -> list[Update]:
        return [
            _get_dataclass_object(data=i, type_=Update)
            for i in self._send_request(
                method="getUpdates",
                parameters={
                    "offset": offset,
                    "limit": limit,
                    "timeout": timeout,
                    "allowed_updates": allowed_updates
                }
            )
        ]

    def set_webhook(
        self,
        url: str,
        *,
        certificate: Optional[InputFile] = None,
        ip_address: Optional[str] = None,
        max_connections: Optional[int] = None,
        allowed_updates: Optional[list[str]] = None,
        drop_pending_updates: Optional[bool] = None,
        secret_token: Optional[str] = None
    ) -> Literal[True]:
        return self._send_request(
            method="setWebhook",
            parameters={
                "url": url,
                "certificate": certificate,
                "ip_address": ip_address,
                "max_connections": max_connections,
                "allowed_updates": allowed_updates,
                "drop_pending_updates": drop_pending_updates,
                "secret_token": secret_token
            }
        )

    def delete_webhook(self, *, drop_pending_updates: Optional[bool] = None) -> Literal[True]:
        return self._send_request(
            method="deleteWebhook",
            parameters={
                "drop_pending_updates": drop_pending_updates
            }
        )

    def get_webhook_info(self) -> WebhookInfo:
        return _get_dataclass_object(
            data=self._send_request(method="getWebhookInfo"),
            type_=WebhookInfo
        )

    def get_me(self) -> User:
        return _get_dataclass_object(
            data=self._send_request(method="getMe"),
            type_=User
        )

    def log_out(self) -> Literal[True]:
        return self._send_request(method="logOut")

    def close(self) -> Literal[True]:
        return self._send_request(method="close")

    def send_message(
        self,
        chat_id: Union[int, str],
        text: str,
        *,
        parse_mode: Optional[str] = None,
        entities: Optional[list[MessageEntity]] = None,
        disable_web_page_preview: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendMessage",
                parameters={
                    "chat_id": chat_id,
                    "text": text,
                    "parse_mode": parse_mode,
                    "entities": entities,
                    "disable_web_page_preview": disable_web_page_preview,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def forward_message(
        self,
        chat_id: Union[int, str],
        from_chat_id: Union[int, str],
        message_id: int,
        *,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="forwardMessage",
                parameters={
                    "chat_id": chat_id,
                    "from_chat_id": from_chat_id,
                    "message_id": message_id,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content
                }
            ),
            type_=Message
        )

    def copy_message(
        self,
        chat_id: Union[int, str],
        from_chat_id: Union[int, str],
        message_id: int,
        *,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> MessageId:
        return _get_dataclass_object(
            data=self._send_request(
                method="copyMessage",
                parameters={
                    "chat_id": chat_id,
                    "from_chat_id": from_chat_id,
                    "message_id": message_id,
                    "caption": caption,
                    "parse_mode": parse_mode,
                    "caption_entities": caption_entities,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=MessageId
        )

    def send_photo(
        self,
        chat_id: Union[int, str],
        photo: Union[InputFile, str],
        *,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendPhoto",
                parameters={
                    "chat_id": chat_id,
                    "photo": photo,
                    "caption": caption,
                    "parse_mode": parse_mode,
                    "caption_entities": caption_entities,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_audio(
        self,
        chat_id: Union[int, str],
        audio: Union[InputFile, str],
        *,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        duration: Optional[int] = None,
        performer: Optional[str] = None,
        title: Optional[str] = None,
        thumb: Union[InputFile, str, None] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendAudio",
                parameters={
                    "chat_id": chat_id,
                    "audio": audio,
                    "caption": caption,
                    "parse_mode": parse_mode,
                    "caption_entities": caption_entities,
                    "duration": duration,
                    "performer": performer,
                    "title": title,
                    "thumb": thumb,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_document(
        self,
        chat_id: Union[int, str],
        document: Union[InputFile, str],
        *,
        thumb: Union[InputFile, str, None] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        disable_content_type_detection: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendDocument",
                parameters={
                    "chat_id": chat_id,
                    "document": document,
                    "thumb": thumb,
                    "caption": caption,
                    "parse_mode": parse_mode,
                    "caption_entities": caption_entities,
                    "disable_content_type_detection": disable_content_type_detection,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_video(
        self,
        chat_id: Union[int, str],
        video: Union[InputFile, str],
        *,
        duration: Optional[int] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        thumb: Union[InputFile, str, None] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        supports_streaming: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendVideo",
                parameters={
                    "chat_id": chat_id,
                    "video": video,
                    "duration": duration,
                    "width": width,
                    "height": height,
                    "thumb": thumb,
                    "caption": caption,
                    "parse_mode": parse_mode,
                    "caption_entities": caption_entities,
                    "supports_streaming": supports_streaming,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_animation(
        self,
        chat_id: Union[int, str],
        animation: Union[InputFile, str],
        *,
        duration: Optional[int] = None,
        width: Optional[int] = None,
        height: Optional[int] = None,
        thumb: Union[InputFile, str, None] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendAnimation",
                parameters={
                    "chat_id": chat_id,
                    "animation": animation,
                    "duration": duration,
                    "width": width,
                    "height": height,
                    "thumb": thumb,
                    "caption": caption,
                    "parse_mode": parse_mode,
                    "caption_entities": caption_entities,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_voice(
        self,
        chat_id: Union[int, str],
        voice: Union[InputFile, str],
        *,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        duration: Optional[int] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendVoice",
                parameters={
                    "chat_id": chat_id,
                    "voice": voice,
                    "caption": caption,
                    "parse_mode": parse_mode,
                    "caption_entities": caption_entities,
                    "duration": duration,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_video_note(
        self,
        chat_id: Union[int, str],
        video_note: Union[InputFile, str],
        *,
        duration: Optional[int] = None,
        length: Optional[int] = None,
        thumb: Union[InputFile, str, None] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendVideoNote",
                parameters={
                    "chat_id": chat_id,
                    "video_note": video_note,
                    "duration": duration,
                    "length": length,
                    "thumb": thumb,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_media_group(
        self,
        chat_id: Union[int, str],
        media: Union[list[InputMediaAudio],
                     list[InputMediaDocument],
                     list[InputMediaPhoto],
                     list[InputMediaVideo]],
        *,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None
    ) -> list[Message]:
        return [
            _get_dataclass_object(data=i, type_=Message)
            for i in self._send_request(
                method="sendMediaGroup",
                parameters={
                    "chat_id": chat_id,
                    "media": media,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply
                }
            )
        ]

    def send_location(
        self,
        chat_id: Union[int, str],
        latitude: float,
        longitude: float,
        *,
        horizontal_accuracy: Optional[float] = None,
        live_period: Optional[int] = None,
        heading: Optional[int] = None,
        proximity_alert_radius: Optional[int] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendLocation",
                parameters={
                    "chat_id": chat_id,
                    "latitude": latitude,
                    "longitude": longitude,
                    "horizontal_accuracy": horizontal_accuracy,
                    "live_period": live_period,
                    "heading": heading,
                    "proximity_alert_radius": proximity_alert_radius,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def edit_message_live_location(
        self,
        latitude: float,
        longitude: float,
        *,
        chat_id: Union[int, str, None] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        horizontal_accuracy: Optional[float] = None,
        heading: Optional[int] = None,
        proximity_alert_radius: Optional[int] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, Literal[True]]:
        data = self._send_request(
            method="editMessageLiveLocation",
            parameters={
                "latitude": latitude,
                "longitude": longitude,
                "chat_id": chat_id,
                "message_id": message_id,
                "inline_message_id": inline_message_id,
                "horizontal_accuracy": horizontal_accuracy,
                "heading": heading,
                "proximity_alert_radius": proximity_alert_radius,
                "reply_markup": reply_markup
            }
        )

        return data if data is True else _get_dataclass_object(data=data, type_=Message)

    def stop_message_live_location(
        self,
        *,
        chat_id: Union[int, str, None] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, Literal[True]]:
        data = self._send_request(
            method="stopMessageLiveLocation",
            parameters={
                "chat_id": chat_id,
                "message_id": message_id,
                "inline_message_id": inline_message_id,
                "reply_markup": reply_markup
            }
        )

        return data if data is True else _get_dataclass_object(data=data, type_=Message)

    def send_venue(
        self,
        chat_id: Union[int, str],
        latitude: float,
        longitude: float,
        title: str,
        address: str,
        *,
        foursquare_id: Optional[str] = None,
        foursquare_type: Optional[str] = None,
        google_place_id: Optional[str] = None,
        google_place_type: Optional[str] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendVenue",
                parameters={
                    "chat_id": chat_id,
                    "latitude": latitude,
                    "longitude": longitude,
                    "title": title,
                    "address": address,
                    "foursquare_id": foursquare_id,
                    "foursquare_type": foursquare_type,
                    "google_place_id": google_place_id,
                    "google_place_type": google_place_type,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_contact(
        self,
        chat_id: Union[int, str],
        phone_number: str,
        first_name: str,
        *,
        last_name: Optional[str] = None,
        vcard: Optional[str] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendContact",
                parameters={
                    "chat_id": chat_id,
                    "phone_number": phone_number,
                    "first_name": first_name,
                    "last_name": last_name,
                    "vcard": vcard,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_poll(
        self,
        chat_id: Union[int, str],
        question: str,
        options: list[str],
        *,
        is_anonymous: Optional[bool] = None,
        type_: Optional[str] = None,
        allows_multiple_answers: Optional[bool] = None,
        correct_option_id: Optional[int] = None,
        explanation: Optional[str] = None,
        explanation_parse_mode: Optional[str] = None,
        explanation_entities: Optional[list[MessageEntity]] = None,
        open_period: Optional[int] = None,
        close_date: Optional[datetime_] = None,
        is_closed: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendPoll",
                parameters={
                    "chat_id": chat_id,
                    "question": question,
                    "options": options,
                    "is_anonymous": is_anonymous,
                    "type": type_,
                    "allows_multiple_answers": allows_multiple_answers,
                    "correct_option_id": correct_option_id,
                    "explanation": explanation,
                    "explanation_parse_mode": explanation_parse_mode,
                    "explanation_entities": explanation_entities,
                    "open_period": open_period,
                    "close_date": close_date,
                    "is_closed": is_closed,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_dice(
        self,
        chat_id: Union[int, str],
        *,
        emoji: Optional[str] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendDice",
                parameters={
                    "chat_id": chat_id,
                    "emoji": emoji,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def send_chat_action(self, chat_id: Union[int, str], action: str) -> Literal[True]:
        return self._send_request(
            method="sendChatAction",
            parameters={
                "chat_id": chat_id,
                "action": action
            }
        )

    def get_user_profile_photos(
        self,
        user_id: int,
        *,
        offset: Optional[int] = None,
        limit: Optional[int] = None
    ) -> UserProfilePhotos:
        return _get_dataclass_object(
            data=self._send_request(
                method="getUserProfilePhotos",
                parameters={
                    "user_id": user_id,
                    "offset": offset,
                    "limit": limit
                }
            ),
            type_=UserProfilePhotos
        )

    def get_file(self, file_id: str) -> File:
        return _get_dataclass_object(
            data=self._send_request(
                method="getFile",
                parameters={
                    "file_id": file_id
                }
            ),
            type_=File
        )

    def ban_chat_member(
        self,
        chat_id: Union[int, str],
        user_id: int,
        *,
        until_date: Optional[datetime_] = None,
        revoke_messages: Optional[bool] = None
    ) -> Literal[True]:
        return self._send_request(
            method="banChatMember",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id,
                "until_date": until_date,
                "revoke_messages": revoke_messages
            }
        )

    def unban_chat_member(
        self,
        chat_id: Union[int, str],
        user_id: int,
        *,
        only_if_banned: Optional[bool] = None
    ) -> Literal[True]:
        return self._send_request(
            method="unbanChatMember",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id,
                "only_if_banned": only_if_banned
            }
        )

    def restrict_chat_member(
        self,
        chat_id: Union[int, str],
        user_id: int,
        permissions: ChatPermissions,
        *,
        until_date: Optional[datetime_] = None
    ) -> Literal[True]:
        return self._send_request(
            method="restrictChatMember",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id,
                "permissions": permissions,
                "until_date": until_date
            }
        )

    def promote_chat_member(
        self,
        chat_id: Union[int, str],
        user_id: int,
        *,
        is_anonymous: Optional[bool] = None,
        can_manage_chat: Optional[bool] = None,
        can_post_messages: Optional[bool] = None,
        can_edit_messages: Optional[bool] = None,
        can_delete_messages: Optional[bool] = None,
        can_manage_video_chats: Optional[bool] = None,
        can_restrict_members: Optional[bool] = None,
        can_promote_members: Optional[bool] = None,
        can_change_info: Optional[bool] = None,
        can_invite_users: Optional[bool] = None,
        can_pin_messages: Optional[bool] = None
    ) -> Literal[True]:
        return self._send_request(
            method="promoteChatMember",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id,
                "is_anonymous": is_anonymous,
                "can_manage_chat": can_manage_chat,
                "can_post_messages": can_post_messages,
                "can_edit_messages": can_edit_messages,
                "can_delete_messages": can_delete_messages,
                "can_manage_video_chats": can_manage_video_chats,
                "can_restrict_members": can_restrict_members,
                "can_promote_members": can_promote_members,
                "can_change_info": can_change_info,
                "can_invite_users": can_invite_users,
                "can_pin_messages": can_pin_messages
            }
        )

    def set_chat_administrator_custom_title(
        self,
        chat_id: Union[int, str],
        user_id: int,
        custom_title: str
    ) -> Literal[True]:
        return self._send_request(
            method="setChatAdministratorCustomTitle",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id,
                "custom_title": custom_title
            }
        )

    def ban_chat_sender_chat(
        self,
        chat_id: Union[int, str],
        sender_chat_id: int
    ) -> Literal[True]:
        return self._send_request(
            method="banChatSenderChat",
            parameters={
                "chat_id": chat_id,
                "sender_chat_id": sender_chat_id
            }
        )

    def unban_chat_sender_chat(
        self,
        chat_id: Union[int, str],
        sender_chat_id: int
    ) -> Literal[True]:
        return self._send_request(
            method="unbanChatSenderChat",
            parameters={
                "chat_id": chat_id,
                "sender_chat_id": sender_chat_id
            }
        )

    def set_chat_permissions(
        self,
        chat_id: Union[int, str],
        permissions: ChatPermissions
    ) -> Literal[True]:
        return self._send_request(
            method="setChatPermissions",
            parameters={
                "chat_id": chat_id,
                "permissions": permissions
            }
        )

    def export_chat_invite_link(self, chat_id: Union[int, str]) -> str:
        return self._send_request(
            method="exportChatInviteLink",
            parameters={
                "chat_id": chat_id
            }
        )

    def create_chat_invite_link(
        self,
        chat_id: Union[int, str],
        *,
        name: Optional[str] = None,
        expire_date: Optional[datetime_] = None,
        member_limit: Optional[int] = None,
        creates_join_request: Optional[bool] = None
    ) -> ChatInviteLink:
        return _get_dataclass_object(
            data=self._send_request(
                method="createChatInviteLink",
                parameters={
                    "chat_id": chat_id,
                    "name": name,
                    "expire_date": expire_date,
                    "member_limit": member_limit,
                    "creates_join_request": creates_join_request
                }
            ),
            type_=ChatInviteLink
        )

    def edit_chat_invite_link(
        self,
        chat_id: Union[int, str],
        invite_link: str,
        *,
        name: Optional[str] = None,
        expire_date: Optional[datetime_] = None,
        member_limit: Optional[int] = None,
        creates_join_request: Optional[bool] = None
    ) -> ChatInviteLink:
        return _get_dataclass_object(
            data=self._send_request(
                method="editChatInviteLink",
                parameters={
                    "chat_id": chat_id,
                    "invite_link": invite_link,
                    "name": name,
                    "expire_date": expire_date,
                    "member_limit": member_limit,
                    "creates_join_request": creates_join_request
                }
            ),
            type_=ChatInviteLink
        )

    def revoke_chat_invite_link(
        self,
        chat_id: Union[int, str],
        invite_link: str
    ) -> ChatInviteLink:
        return _get_dataclass_object(
            data=self._send_request(
                method="revokeChatInviteLink",
                parameters={
                    "chat_id": chat_id,
                    "invite_link": invite_link
                }
            ),
            type_=ChatInviteLink
        )

    def approve_chat_join_request(self, chat_id: Union[int, str], user_id: int) -> Literal[True]:
        return self._send_request(
            method="approveChatJoinRequest",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id
            }
        )

    def decline_chat_join_request(self, chat_id: Union[int, str], user_id: int) -> Literal[True]:
        return self._send_request(
            method="declineChatJoinRequest",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id
            }
        )

    def set_chat_photo(self, chat_id: Union[int, str], photo: InputFile) -> Literal[True]:
        return self._send_request(
            method="setChatPhoto",
            parameters={
                "chat_id": chat_id,
                "photo": photo
            }
        )

    def delete_chat_photo(self, chat_id: Union[int, str]) -> Literal[True]:
        return self._send_request(
            method="deleteChatPhoto",
            parameters={
                "chat_id": chat_id
            }
        )
    
    def set_chat_title(self, chat_id: Union[int, str], title: str) -> Literal[True]:
        return self._send_request(
            method="setChatTitle",
            parameters={
                "chat_id": chat_id,
                "title": title
            }
        )

    def set_chat_description(self, chat_id: Union[int, str], description: str) -> Literal[True]:
        return self._send_request(
            method="setChatDescription",
            parameters={
                "chat_id": chat_id,
                "description": description
            }
        )

    def pin_chat_message(
        self,
        chat_id: Union[int, str],
        message_id: int,
        *,
        disable_notification: Optional[bool] = None
    ) -> Literal[True]:
        return self._send_request(
            method="pinChatMessage",
            parameters={
                "chat_id": chat_id,
                "message_id": message_id,
                "disable_notification": disable_notification
            }
        )

    def unpin_chat_message(self, chat_id: Union[int, str], message_id: int) -> Literal[True]:
        return self._send_request(
            method="unpinChatMessage",
            parameters={
                "chat_id": chat_id,
                "message_id": message_id
            }
        )

    def unpin_all_chat_messages(self, chat_id: Union[int, str]) -> Literal[True]:
        return self._send_request(
            method="unpinAllChatMessages",
            parameters={
                "chat_id": chat_id
            }
        )

    def leave_chat(self, chat_id: Union[int, str]) -> Literal[True]:
        return self._send_request(
            method="leaveChat",
            parameters={
                "chat_id": chat_id
            }
        )

    def get_chat(self, chat_id: Union[int, str]) -> Chat:
        return _get_dataclass_object(
            data=self._send_request(
                method="getChat",
                parameters={
                    "chat_id": chat_id
                }
            ),
            type_=Chat
        )

    def get_chat_administrators(
        self,
        chat_id: Union[int, str]
    ) -> list[Union[ChatMemberOwner,
                    ChatMemberAdministrator]]:
        return [
            _get_dataclass_object(data=i, type_=_CHAT_MEMBER_TYPES[i["status"]])
            for i in self._send_request(
                method="getChatAdministrators",
                parameters={
                    "chat_id": chat_id
                }
            )
        ]

    def get_chat_member_count(self, chat_id: Union[int, str]) -> int:
        return self._send_request(
            method="getChatMemberCount",
            parameters={
                "chat_id": chat_id
            }
        )

    def get_chat_member(self, chat_id: Union[int, str], user_id: int) -> ChatMember:
        data = self._send_request(
            method="getChatMember",
            parameters={
                "chat_id": chat_id,
                "user_id": user_id
            }
        )

        return _get_dataclass_object(data=data, type_=_CHAT_MEMBER_TYPES[data["status"]])

    def set_chat_sticker_set(
        self,
        chat_id: Union[int, str],
        sticker_set_name: str
    ) -> Literal[True]:
        return self._send_request(
            method="setChatStickerSet",
            parameters={
                "chat_id": chat_id,
                "sticker_set_name": sticker_set_name
            }
        )

    def delete_chat_sticker_set(self, chat_id: Union[int, str]) -> Literal[True]:
        return self._send_request(
            method="deleteChatStickerSet",
            parameters={
                "chat_id": chat_id
            }
        )

    def answer_callback_query(
        self,
        callback_query_id: str,
        *,
        text: Optional[str] = None,
        show_alert: Optional[bool] = None,
        url: Optional[str] = None,
        cache_time: Optional[int] = None
    ) -> Literal[True]:
        return self._send_request(
            method="answerCallbackQuery",
            parameters={
                "callback_query_id": callback_query_id,
                "text": text,
                "show_alert": show_alert,
                "url": url,
                "cache_time": cache_time
            }
        )

    def set_my_commands(
        self,
        commands: list[BotCommand],
        *,
        scope: Optional[BotCommandScope] = None,
        language_code: Optional[str] = None
    ) -> Literal[True]:
        return self._send_request(
            method="setMyCommands",
            parameters={
                "commands": commands,
                "scope": scope,
                "language_code": language_code
            }
        )

    def delete_my_commands(
        self,
        *,
        scope: Optional[BotCommandScope] = None,
        language_code: Optional[str] = None
    ) -> Literal[True]:
        return self._send_request(
            method="deleteMyCommands",
            parameters={
                "scope": scope,
                "language_code": language_code
            }
        )

    def get_my_commands(
        self,
        *,
        scope: Optional[BotCommandScope] = None,
        language_code: Optional[str] = None
    ) -> list[BotCommand]:
        return [
            _get_dataclass_object(data=i, type_=BotCommand)
            for i in self._send_request(
                method="getMyCommands",
                parameters={
                    "scope": scope,
                    "language_code": language_code
                }
            )
        ]

    def set_chat_menu_button(
        self,
        *,
        chat_id: Optional[int] = None,
        menu_button: Optional[MenuButton] = None
    ) -> Literal[True]:
        return self._send_request(
            method="setChatMenuButton",
            parameters={
                "chat_id": chat_id,
                "menu_button": menu_button
            }
        )

    def get_chat_menu_button(self, *, chat_id: Optional[int] = None) -> Literal[True]:
        return self._send_request(
            method="getChatMenuButton",
            parameters={
                "chat_id": chat_id
            }
        )

    def set_my_default_administrator_rights(
        self,
        *,
        rights: Optional[ChatAdministratorRights] = None,
        for_channels: Optional[bool] = None
    ) -> Literal[True]:
        return self._send_request(
            method="setMyDefaultAdministratorRights",
            parameters={
                "rights": rights,
                "for_channels": for_channels
            }
        )

    def get_my_default_administrator_rights(
        self,
        *,
        for_channels: Optional[bool] = None
    ) -> Literal[True]:
        return self._send_request(
            method="getMyDefaultAdministratorRights",
            parameters={
                "for_channels": for_channels
            }
        )

    def edit_message_text(
        self,
        text: str,
        *,
        chat_id: Union[int, str, None] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        parse_mode: Optional[str] = None,
        entities: Optional[list[MessageEntity]] = None,
        disable_web_page_preview: Optional[bool] = None,
        reply_markup: Optional[ReplyKeyboardMarkup] = None
    ) -> Union[Message, Literal[True]]:
        data = self._send_request(
            method="editMessageText",
            parameters={
                "text": text,
                "chat_id": chat_id,
                "message_id": message_id,
                "inline_message_id": inline_message_id,
                "parse_mode": parse_mode,
                "entities": entities,
                "disable_web_page_preview": disable_web_page_preview,
                "reply_markup": reply_markup
            }
        )

        return data if data is True else _get_dataclass_object(data=data, type_=Message)

    def edit_message_caption(
        self,
        *,
        chat_id: Union[int, str, None] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        caption: Optional[str] = None,
        parse_mode: Optional[str] = None,
        caption_entities: Optional[list[MessageEntity]] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, Literal[True]]:
        data = self._send_request(
            method="editMessageCaption",
            parameters={
                "chat_id": chat_id,
                "message_id": message_id,
                "inline_message_id": inline_message_id,
                "caption": caption,
                "parse_mode": parse_mode,
                "caption_entities": caption_entities,
                "reply_markup": reply_markup
            }
        )

        return data if data is True else _get_dataclass_object(data=data, type_=Message)

    def edit_message_media(
        self,
        media: InputMedia,
        *,
        chat_id: Union[int, str, None] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, Literal[True]]:
        data = self._send_request(
            method="editMessageMedia",
            parameters={
                "media": media,
                "chat_id": chat_id,
                "message_id": message_id,
                "inline_message_id": inline_message_id,
                "reply_markup": reply_markup
            }
        )

        return data if data is True else _get_dataclass_object(data=data, type_=Message)

    def edit_message_reply_markup(
        self,
        *,
        chat_id: Union[int, str, None] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Union[Message, Literal[True]]:
        data = self._send_request(
            method="editMessageReplyMarkup",
            parameters={
                "chat_id": chat_id,
                "message_id": message_id,
                "inline_message_id": inline_message_id,
                "reply_markup": reply_markup
            }
        )

        return data if data is True else _get_dataclass_object(data=data, type_=Message)

    def stop_poll(
        self,
        chat_id: Union[int, str],
        message_id: int,
        *,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Poll:
        return _get_dataclass_object(
            data=self._send_request(
                method="stopPoll",
                parameters={
                    "chat_id": chat_id,
                    "message_id": message_id,
                    "reply_markup": reply_markup
                }
            ),
            type_=Poll
        )

    def delete_message(self, chat_id: Union[int, str], message_id: int) -> Literal[True]:
        return self._send_request(
            method="deleteMessage",
            parameters={
                "chat_id": chat_id,
                "message_id": message_id
            }
        )

    def send_sticker(
        self,
        chat_id: Union[int, str],
        sticker: Union[InputFile, str],
        *,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Union[InlineKeyboardMarkup,
                            ReplyKeyboardMarkup,
                            ReplyKeyboardRemove,
                            ForceReply,
                            None] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendSticker",
                parameters={
                    "chat_id": chat_id,
                    "sticker": sticker,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def get_sticker_set(self, name: str) -> StickerSet:
        return _get_dataclass_object(
            data=self._send_request(
                method="getStickerSet",
                parameters={
                    "name": name
                }
            ),
            type_=StickerSet
        )

    def get_custom_emoji_stickers(self, custom_emoji_ids: list[str]) -> list[Sticker]:
        return [
            _get_dataclass_object(data=i, type_=Sticker)
            for i in self._send_request(
                method="getCustomEmojiStickers",
                parameters={
                    "custom_emoji_ids": custom_emoji_ids
                }
            )
        ]

    def upload_sticker_file(self, user_id: int, png_sticker: InputFile) -> File:
        return _get_dataclass_object(
            data=self._send_request(
                method="uploadStickerFile",
                parameters={
                    "user_id": user_id,
                    "png_sticker": png_sticker
                }
            ),
            type_=File
        )

    def create_new_sticker_set(
        self,
        user_id: int,
        name: str,
        title: str,
        emojis: str,
        *,
        png_sticker: Union[InputFile, str, None] = None,
        tgs_sticker: Optional[InputFile] = None,
        webm_sticker: Optional[InputFile] = None,
        contains_masks: Optional[bool] = None,
        mask_position: Optional[MaskPosition] = None
    ) -> Literal[True]:
        return self._send_request(
            method="createNewStickerSet",
            parameters={
                "user_id": user_id,
                "name": name,
                "title": title,
                "emojis": emojis,
                "png_sticker": png_sticker,
                "tgs_sticker": tgs_sticker,
                "webm_sticker": webm_sticker,
                "contains_masks": contains_masks,
                "mask_position": mask_position
            }
        )

    def add_sticker_to_set(
        self,
        user_id: int,
        name: str,
        emojis: str,
        *,
        png_sticker: Union[InputFile, str, None] = None,
        tgs_sticker: Optional[InputFile] = None,
        webm_sticker: Optional[InputFile] = None,
        mask_position: Optional[MaskPosition] = None
    ) -> Literal[True]:
        return self._send_request(
            method="addStickerToSet",
            parameters={
                "user_id": user_id,
                "name": name,
                "emojis": emojis,
                "png_sticker": png_sticker,
                "tgs_sticker": tgs_sticker,
                "webm_sticker": webm_sticker,
                "mask_position": mask_position
            }
        )

    def set_sticker_position_in_set(self, sticker: str, position: int) -> Literal[True]:
        return self._send_request(
            method="setStickerPositionInSet",
            parameters={
                "sticker": sticker,
                "position": position
            }
        )

    def delete_sticker_from_set(self, sticker: str) -> Literal[True]:
        return self._send_request(
            method="deleteStickerFromSet",
            parameters={
                "sticker": sticker
            }
        )

    def set_sticker_set_thumb(
        self,
        name: str,
        user_id: int,
        *,
        thumb: Union[InputFile, str, None] = None
    ) -> Literal[True]:
        return self._send_request(
            method="setStickerSetThumb",
            parameters={
                "name": name,
                "user_id": user_id,
                "thumb": thumb
            }
        )

    def answer_inline_query(
        self,
        inline_query_id: str,
        results: list[InlineQueryResult],
        *,
        cache_time: Optional[int] = None,
        is_personal: Optional[bool] = None,
        next_offset: Optional[str] = None,
        switch_pm_text: Optional[str] = None,
        switch_pm_parameter: Optional[str] = None
    ) -> Literal[True]:
        return self._send_request(
            method="answerInlineQuery",
            parameters={
                "inline_query_id": inline_query_id,
                "results": results,
                "cache_time": cache_time,
                "is_personal": is_personal,
                "next_offset": next_offset,
                "switch_pm_text": switch_pm_text,
                "switch_pm_parameter": switch_pm_parameter
            }
        )

    def answer_web_app_query(
        self,
        web_app_query_id: str,
        result: InlineQueryResult
    ) -> SentWebAppMessage:
        return _get_dataclass_object(
            data=self._send_request(
                method="answerWebAppQuery",
                parameters={
                    "web_app_query_id": web_app_query_id,
                    "result": result
                }
            ),
            type_=SentWebAppMessage
        )

    def send_invoice(
        self,
        chat_id: Union[int, str],
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        currency: str,
        prices: list[LabeledPrice],
        *,
        max_tip_amount: Optional[int] = None,
        suggested_tip_amounts: Optional[list[int]] = None,
        start_parameter: Optional[str] = None,
        provider_data: Optional[str] = None,
        photo_url: Optional[str] = None,
        photo_size: Optional[int] = None,
        photo_width: Optional[int] = None,
        photo_height: Optional[int] = None,
        need_name: Optional[bool] = None,
        need_phone_number: Optional[bool] = None,
        need_email: Optional[bool] = None,
        need_shipping_address: Optional[bool] = None,
        send_phone_number_to_provider: Optional[bool] = None,
        send_email_to_provider: Optional[bool] = None,
        is_flexible: Optional[bool] = None,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendInvoice",
                parameters={
                    "chat_id": chat_id,
                    "title": title,
                    "description": description,
                    "payload": payload,
                    "provider_token": provider_token,
                    "currency": currency,
                    "prices": prices,
                    "max_tip_amount": max_tip_amount,
                    "suggested_tip_amounts": suggested_tip_amounts,
                    "start_parameter": start_parameter,
                    "provider_data": provider_data,
                    "photo_url": photo_url,
                    "photo_size": photo_size,
                    "photo_width": photo_width,
                    "photo_height": photo_height,
                    "need_name": need_name,
                    "need_phone_number": need_phone_number,
                    "need_email": need_email,
                    "need_shipping_address": need_shipping_address,
                    "send_phone_number_to_provider": send_phone_number_to_provider,
                    "send_email_to_provider": send_email_to_provider,
                    "is_flexible": is_flexible,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def create_invoice_link(
        self,
        title: str,
        description: str,
        payload: str,
        provider_token: str,
        currency: str,
        prices: list[LabeledPrice],
        *,
        max_tip_amount: Optional[int] = None,
        suggested_tip_amounts: Optional[list[int]] = None,
        provider_data: Optional[str] = None,
        photo_url: Optional[str] = None,
        photo_size: Optional[int] = None,
        photo_width: Optional[int] = None,
        photo_height: Optional[int] = None,
        need_name: Optional[bool] = None,
        need_phone_number: Optional[bool] = None,
        need_email: Optional[bool] = None,
        need_shipping_address: Optional[bool] = None,
        send_phone_number_to_provider: Optional[bool] = None,
        send_email_to_provider: Optional[bool] = None,
        is_flexible: Optional[bool] = None
    ) -> str:
        return self._send_request(
            method="createInvoiceLink",
            parameters={
                "title": title,
                "description": description,
                "payload": payload,
                "provider_token": provider_token,
                "currency": currency,
                "prices": prices,
                "max_tip_amount": max_tip_amount,
                "suggested_tip_amounts": suggested_tip_amounts,
                "provider_data": provider_data,
                "photo_url": photo_url,
                "photo_size": photo_size,
                "photo_width": photo_width,
                "photo_height": photo_height,
                "need_name": need_name,
                "need_phone_number": need_phone_number,
                "need_email": need_email,
                "need_shipping_address": need_shipping_address,
                "send_phone_number_to_provider": send_phone_number_to_provider,
                "send_email_to_provider": send_email_to_provider,
                "is_flexible": is_flexible
            }
        )

    def answer_shipping_query(
        self,
        shipping_query_id: str,
        ok: bool,
        *,
        shipping_options: Optional[list[ShippingOption]] = None,
        error_message: Optional[str] = None
    ) -> Literal[True]:
        return self._send_request(
            method="answerShippingQuery",
            parameters={
                "shipping_query_id": shipping_query_id,
                "ok": ok,
                "shipping_options": shipping_options,
                "error_message": error_message
            }
        )

    def answer_pre_checkout_query(
        self,
        pre_checkout_query_id: str,
        ok: bool,
        *,
        error_message: Optional[str] = None
    ) -> Literal[True]:
        return self._send_request(
            method="answerPreCheckoutQuery",
            parameters={
                "pre_checkout_query_id": pre_checkout_query_id,
                "ok": ok,
                "error_message": error_message
            }
        )

    def set_passport_data_errors(
        self,
        user_id: int,
        errors: list[PassportElementError]
    ) -> Literal[True]:
        return self._send_request(
            method="setPassportDataErrors",
            parameters={
                "user_id": user_id,
                "errors": errors
            }
        )

    def send_game(
        self,
        chat_id: int,
        game_short_name: str,
        *,
        disable_notification: Optional[bool] = None,
        protect_content: Optional[bool] = None,
        reply_to_message_id: Optional[int] = None,
        allow_sending_without_reply: Optional[bool] = None,
        reply_markup: Optional[InlineKeyboardMarkup] = None
    ) -> Message:
        return _get_dataclass_object(
            data=self._send_request(
                method="sendGame",
                parameters={
                    "chat_id": chat_id,
                    "game_short_name": game_short_name,
                    "disable_notification": disable_notification,
                    "protect_content": protect_content,
                    "reply_to_message_id": reply_to_message_id,
                    "allow_sending_without_reply": allow_sending_without_reply,
                    "reply_markup": reply_markup
                }
            ),
            type_=Message
        )

    def set_game_score(
        self,
        user_id: int,
        score: int,
        *,
        force: Optional[bool] = None,
        disable_edit_message: Optional[bool] = None,
        chat_id: Optional[int] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None
    ) -> Union[Message, Literal[True]]:
        data = self._send_request(
            method="setGameScore",
            parameters={
                "user_id": user_id,
                "score": score,
                "force": force,
                "disable_edit_message": disable_edit_message,
                "chat_id": chat_id,
                "message_id": message_id,
                "inline_message_id": inline_message_id
            }
        )

        return data if data is True else _get_dataclass_object(data=data, type_=Message)

    def get_game_high_scores(
        self,
        user_id: int,
        *,
        chat_id: Optional[int] = None,
        message_id: Optional[int] = None,
        inline_message_id: Optional[str] = None
    ) -> list[GameHighScore]:
        return [
            _get_dataclass_object(data=i, type_=GameHighScore)
            for i in self._send_request(
                method="getGameHighScores",
                parameters={
                    "user_id": user_id,
                    "chat_id": chat_id,
                    "message_id": message_id,
                    "inline_message_id": inline_message_id
                }
            )
        ]

    def _send_request(
        self,
        method: str,
        *,
        parameters: Optional[dict[str, Any]] = None
    ) -> Any:
        parameters = parameters or {}
        url = self._get_url(method)
        data, files = _get_prepared_parameters(parameters)
        response = self._session.post(url, json=data, files=files)

        return _process_response(response, method, parameters)

    def _get_url(self, method: str) -> str:
        return f"{self._url}/bot{self._token}/{method}"


def _get_prepared_parameters(
    parameters: dict[str, Any]
) -> tuple[dict[str, Any], list[tuple[str, tuple[str, bytes]]]]:
    data = {}
    files = []

    for name, value in parameters.items():
        if value is not None:
            if isinstance(value, InputFile):
                files.append((name, (value.name or secrets.token_urlsafe(16), value.content)))
            else:
                if isinstance(value, datetime_):
                    value = _convert_datetime_to_timestamp(value)
                elif is_dataclass(value):
                    value = _get_dataclass_object_compressed_data(value)

                data[name] = value

    return data, files


def _process_response(response: Response, method: str, parameters: dict[str, Any]) -> Any:
    data = response.json()

    if not data["ok"]:
        try:
            response_parameter_data = data["parameters"]
        except KeyError:
            response_parameters = None
        else:
            response_parameters = _get_dataclass_object(
                data=response_parameter_data,
                type_=ResponseParameters
            )

        raise get_error(
            method=method,
            parameters=parameters,
            status_code=response.status_code,
            description=data["description"],
            response_parameters=response_parameters
        )

    return data["result"]


def _get_dataclass_object_compressed_data(object_: Any) -> dict[str, Any]:
    data = _dataclass_factory.dump(object_)
    _compress_dict_data(data)

    return data


def _get_dataclass_object(data: dict[str, Any], type_: Type[DataclassObject]) -> DataclassObject:
    return _dataclass_factory.load(data, type_)


def _compress_dict_data(data: dict[str, Any]) -> None:
    for key in tuple(data):
        if data[key] is None:
            del data[key]
        elif isinstance(data[key], dict):
            _compress_dict_data(data[key])
        elif isinstance(data[key], list):
            _compress_list_data(data[key])


def _compress_list_data(data: list[Any]) -> None:
    for i in data:
        if isinstance(i, dict):
            _compress_dict_data(i)
        elif isinstance(i, list):
            _compress_list_data(i)



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.user import User
from telebox.telegram.types.types.shipping_address import ShippingAddress


@dataclass(unsafe_hash=True)
class ShippingQuery(Type):
    id: str
    from_: User
    invoice_payload: str
    shipping_address: ShippingAddress



from dataclasses import dataclass
from typing import Literal, Optional, TYPE_CHECKING

from telebox.telegram.types.base import Type
from telebox.telegram.utils import get_url, get_full_name, get_chat_id_without_prefix
from telebox.telegram.consts import chat_types
from telebox.telegram.types.types.chat_photo import ChatPhoto
from telebox.telegram.types.types.chat_permissions import ChatPermissions
from telebox.telegram.types.types.chat_location import ChatLocation
if TYPE_CHECKING:
    from telebox.telegram.types.types.message import Message


@dataclass(unsafe_hash=True)
class Chat(Type):
    id: int
    type: str
    title: Optional[str] = None
    username: Optional[str] = None
    first_name: Optional[str] = None
    last_name: Optional[str] = None
    photo: Optional[ChatPhoto] = None
    bio: Optional[str] = None
    has_private_forwards: Optional[Literal[True]] = None
    has_restricted_voice_and_video_messages: Optional[Literal[True]] = None
    join_to_send_messages: Optional[Literal[True]] = None
    join_by_request: Optional[Literal[True]] = None
    description: Optional[str] = None
    invite_link: Optional[str] = None
    pinned_message: Optional["Message"] = None
    permissions: Optional[ChatPermissions] = None
    slow_mode_delay: Optional[int] = None
    message_auto_delete_time: Optional[int] = None
    has_protected_content: Optional[Literal[True]] = None
    sticker_set_name: Optional[str] = None
    can_set_sticker_set: Optional[Literal[True]] = None
    linked_chat_id: Optional[int] = None
    location: Optional[ChatLocation] = None

    @property
    def id_without_prefix(self) -> int:
        if self.type in {chat_types.CHANNEL, chat_types.SUPERGROUP}:
            return get_chat_id_without_prefix(self.id)

        return self.id

    @property
    def full_name(self) -> Optional[str]:
        if self.first_name is not None:
            return get_full_name(self.first_name, self.last_name)

    @property
    def url(self) -> Optional[str]:
        if self.username is not None:
            return get_url(self.username)
        elif self.invite_link is not None:
            return self.invite_link



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.consts import chat_member_statuses
from telebox.telegram.types.types.user import User


@dataclass(unsafe_hash=True)
class ChatMemberBanned(Type):
    user: User
    until_date: Optional[datetime] = None  # None instead of 0
    status: str = chat_member_statuses.KICKED



from .types.update import Update
from .types.webhook_info import WebhookInfo
from .types.user import User
from .types.chat import Chat
from .types.message import Message
from .types.message_id import MessageId
from .types.message_entity import MessageEntity
from .types.photo_size import PhotoSize
from .types.animation import Animation
from .types.audio import Audio
from .types.document import Document
from .types.video import Video
from .types.video_note import VideoNote
from .types.voice import Voice
from .types.contact import Contact
from .types.dice import Dice
from .types.poll_option import PollOption
from .types.poll_answer import PollAnswer
from .types.poll import Poll
from .types.location import Location
from .types.venue import Venue
from .types.web_app_data import WebAppData
from .types.proximity_alert_triggered import ProximityAlertTriggered
from .types.message_auto_delete_timer_changed import MessageAutoDeleteTimerChanged
from .types.video_chat_scheduled import VideoChatScheduled
from .types.video_chat_started import VideoChatStarted
from .types.video_chat_ended import VideoChatEnded
from .types.video_chat_participants_invited import VideoChatParticipantsInvited
from .types.user_profile_photos import UserProfilePhotos
from .types.file import File
from .types.web_app_info import WebAppInfo
from .types.reply_keyboard_markup import ReplyKeyboardMarkup
from .types.keyboard_button import KeyboardButton
from .types.keyboard_button_poll_type import KeyboardButtonPollType
from .types.reply_keyboard_remove import ReplyKeyboardRemove
from .types.inline_keyboard_markup import InlineKeyboardMarkup
from .types.inline_keyboard_button import InlineKeyboardButton
from .types.login_url import LoginUrl
from .types.callback_query import CallbackQuery
from .types.force_reply import ForceReply
from .types.chat_photo import ChatPhoto
from .types.chat_invite_link import ChatInviteLink
from .types.chat_administrator_rights import ChatAdministratorRights
from .types.chat_member import ChatMember
from .types.chat_member_owner import ChatMemberOwner
from .types.chat_member_administrator import ChatMemberAdministrator
from .types.chat_member_member import ChatMemberMember
from .types.chat_member_restricted import ChatMemberRestricted
from .types.chat_member_left import ChatMemberLeft
from .types.chat_member_banned import ChatMemberBanned
from .types.chat_member_updated import ChatMemberUpdated
from .types.chat_join_request import ChatJoinRequest
from .types.chat_permissions import ChatPermissions
from .types.chat_location import ChatLocation
from .types.bot_command import BotCommand
from .types.bot_command_scope import BotCommandScope
from .types.bot_command_scope_default import BotCommandScopeDefault
from .types.bot_command_scope_all_private_chats import BotCommandScopeAllPrivateChats
from .types.bot_command_scope_all_group_chats import BotCommandScopeAllGroupChats
from .types.bot_command_scope_all_chat_administrators import BotCommandScopeAllChatAdministrators
from .types.bot_command_scope_chat import BotCommandScopeChat
from .types.bot_command_scope_chat_administrators import BotCommandScopeChatAdministrators
from .types.bot_command_scope_chat_member import BotCommandScopeChatMember
from .types.menu_button import MenuButton
from .types.menu_button_commands import MenuButtonCommands
from .types.menu_button_web_app import MenuButtonWebApp
from .types.menu_button_default import MenuButtonDefault
from .types.response_parameters import ResponseParameters
from .types.input_media import InputMedia
from .types.input_media_photo import InputMediaPhoto
from .types.input_media_video import InputMediaVideo
from .types.input_media_animation import InputMediaAnimation
from .types.input_media_audio import InputMediaAudio
from .types.input_media_document import InputMediaDocument
from .types.input_file import InputFile
from .types.sticker import Sticker
from .types.sticker_set import StickerSet
from .types.mask_position import MaskPosition
from .types.inline_query import InlineQuery
from .types.inline_query_result import InlineQueryResult
from .types.inline_query_result_article import InlineQueryResultArticle
from .types.inline_query_result_photo import InlineQueryResultPhoto
from .types.inline_query_result_gif import InlineQueryResultGif
from .types.inline_query_result_mpeg4_gif import InlineQueryResultMpeg4Gif
from .types.inline_query_result_video import InlineQueryResultVideo
from .types.inline_query_result_audio import InlineQueryResultAudio
from .types.inline_query_result_voice import InlineQueryResultVoice
from .types.inline_query_result_document import InlineQueryResultDocument
from .types.inline_query_result_location import InlineQueryResultLocation
from .types.inline_query_result_venue import InlineQueryResultVenue
from .types.inline_query_result_contact import InlineQueryResultContact
from .types.inline_query_result_game import InlineQueryResultGame
from .types.inline_query_result_cached_photo import InlineQueryResultCachedPhoto
from .types.inline_query_result_cached_gif import InlineQueryResultCachedGif
from .types.inline_query_result_cached_mpeg4_gif import InlineQueryResultCachedMpeg4Gif
from .types.inline_query_result_cached_sticker import InlineQueryResultCachedSticker
from .types.inline_query_result_cached_document import InlineQueryResultCachedDocument
from .types.inline_query_result_cached_video import InlineQueryResultCachedVideo
from .types.inline_query_result_cached_voice import InlineQueryResultCachedVoice
from .types.inline_query_result_cached_audio import InlineQueryResultCachedAudio
from .types.input_message_content import InputMessageContent
from .types.input_text_message_content import InputTextMessageContent
from .types.input_location_message_content import InputLocationMessageContent
from .types.input_venue_message_content import InputVenueMessageContent
from .types.input_contact_message_content import InputContactMessageContent
from .types.input_invoice_message_content import InputInvoiceMessageContent
from .types.chosen_inline_result import ChosenInlineResult
from .types.sent_web_app_message import SentWebAppMessage
from .types.labeled_price import LabeledPrice
from .types.invoice import Invoice
from .types.shipping_address import ShippingAddress
from .types.order_info import OrderInfo
from .types.shipping_option import ShippingOption
from .types.successful_payment import SuccessfulPayment
from .types.shipping_query import ShippingQuery
from .types.pre_checkout_query import PreCheckoutQuery
from .types.passport_data import PassportData
from .types.passport_file import PassportFile
from .types.encrypted_passport_element import EncryptedPassportElement
from .types.encrypted_credentials import EncryptedCredentials
from .types.passport_element_error import PassportElementError
from .types.passport_element_error_data_field import PassportElementErrorDataField
from .types.passport_element_error_front_side import PassportElementErrorFrontSide
from .types.passport_element_error_reverse_side import PassportElementErrorReverseSide
from .types.passport_element_error_selfie import PassportElementErrorSelfie
from .types.passport_element_error_file import PassportElementErrorFile
from .types.passport_element_error_files import PassportElementErrorFiles
from .types.passport_element_error_translation_file import PassportElementErrorTranslationFile
from .types.passport_element_error_translation_files import PassportElementErrorTranslationFiles
from .types.passport_element_error_unspecified import PassportElementErrorUnspecified
from .types.game import Game
from .types.callback_game import CallbackGame
from .types.game_high_score import GameHighScore



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, Literal, TYPE_CHECKING

from telebox.telegram.types.base import Type
from telebox.telegram.utils import get_message_public_url, get_message_private_url
from telebox.telegram.consts import chat_types
from telebox.telegram import enums
from telebox.telegram.errors import UnknownMessageTypeError
from telebox.telegram.types.types.user import User
from telebox.telegram.types.types.message_entity import MessageEntity
from telebox.telegram.types.types.animation import Animation
from telebox.telegram.types.types.audio import Audio
from telebox.telegram.types.types.document import Document
from telebox.telegram.types.types.photo_size import PhotoSize
from telebox.telegram.types.types.sticker import Sticker
from telebox.telegram.types.types.video import Video
from telebox.telegram.types.types.video_note import VideoNote
from telebox.telegram.types.types.voice import Voice
from telebox.telegram.types.types.contact import Contact
from telebox.telegram.types.types.dice import Dice
from telebox.telegram.types.types.game import Game
from telebox.telegram.types.types.poll import Poll
from telebox.telegram.types.types.venue import Venue
from telebox.telegram.types.types.location import Location
from telebox.telegram.types.types.invoice import Invoice
from telebox.telegram.types.types.successful_payment import SuccessfulPayment
from telebox.telegram.types.types.passport_data import PassportData
from telebox.telegram.types.types.proximity_alert_triggered import ProximityAlertTriggered
from telebox.telegram.types.types.video_chat_scheduled import VideoChatScheduled
from telebox.telegram.types.types.video_chat_started import VideoChatStarted
from telebox.telegram.types.types.video_chat_ended import VideoChatEnded
from telebox.telegram.types.types.web_app_data import WebAppData
from telebox.telegram.types.types.inline_keyboard_markup import InlineKeyboardMarkup
from telebox.telegram.types.types.message_auto_delete_timer_changed import (
    MessageAutoDeleteTimerChanged
)
from telebox.telegram.types.types.video_chat_participants_invited import (
    VideoChatParticipantsInvited
)
if TYPE_CHECKING:
    from telebox.telegram.types.types.chat import Chat


@dataclass(unsafe_hash=True)
class Message(Type):
    message_id: int
    date: datetime
    chat: "Chat"
    from_: Optional[User] = None
    sender_chat: Optional["Chat"] = None
    forward_from: Optional[User] = None
    forward_from_chat: Optional["Chat"] = None
    forward_from_message_id: Optional[int] = None
    forward_signature: Optional[str] = None
    forward_sender_name: Optional[str] = None
    forward_date: Optional[datetime] = None
    is_automatic_forward: Optional[Literal[True]] = None
    reply_to_message: Optional["Message"] = None
    via_bot: Optional[User] = None
    edit_date: Optional[datetime] = None
    has_protected_content: Optional[Literal[True]] = None
    media_group_id: Optional[str] = None
    author_signature: Optional[str] = None
    text: Optional[str] = None
    entities: Optional[list[MessageEntity]] = None
    animation: Optional[Animation] = None
    audio: Optional[Audio] = None
    document: Optional[Document] = None
    photo: Optional[list[PhotoSize]] = None
    sticker: Optional[Sticker] = None
    video: Optional[Video] = None
    video_note: Optional[VideoNote] = None
    voice: Optional[Voice] = None
    caption: Optional[str] = None
    caption_entities: Optional[list[MessageEntity]] = None
    contact: Optional[Contact] = None
    dice: Optional[Dice] = None
    game: Optional[Game] = None
    poll: Optional[Poll] = None
    venue: Optional[Venue] = None
    location: Optional[Location] = None
    new_chat_members: Optional[list[User]] = None
    left_chat_member: Optional[User] = None
    new_chat_title: Optional[str] = None
    new_chat_photo: Optional[list[PhotoSize]] = None
    delete_chat_photo: Optional[Literal[True]] = None
    group_chat_created: Optional[Literal[True]] = None
    supergroup_chat_created: Optional[Literal[True]] = None
    channel_chat_created: Optional[Literal[True]] = None
    message_auto_delete_timer_changed: Optional[MessageAutoDeleteTimerChanged] = None
    migrate_to_chat_id: Optional[int] = None
    migrate_from_chat_id: Optional[int] = None
    pinned_message: Optional["Message"] = None
    invoice: Optional[Invoice] = None
    successful_payment: Optional[SuccessfulPayment] = None
    connected_website: Optional[str] = None
    passport_data: Optional[PassportData] = None
    proximity_alert_triggered: Optional[ProximityAlertTriggered] = None
    video_chat_scheduled: Optional[VideoChatScheduled] = None
    video_chat_started: Optional[VideoChatStarted] = None
    video_chat_ended: Optional[VideoChatEnded] = None
    video_chat_participants_invited: Optional[VideoChatParticipantsInvited] = None
    web_app_data: Optional[WebAppData] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None

    @property
    def type(self) -> enums.MessageType:
        if self.text is not None:
            return enums.MessageType.TEXT
        elif self.animation is not None:
            return enums.MessageType.ANIMATION
        elif self.audio is not None:
            return enums.MessageType.AUDIO
        elif self.document is not None:
            return enums.MessageType.DOCUMENT
        elif self.photo is not None:
            return enums.MessageType.PHOTO
        elif self.sticker is not None:
            return enums.MessageType.STICKER
        elif self.video is not None:
            return enums.MessageType.VIDEO
        elif self.video_note is not None:
            return enums.MessageType.VIDEO_NOTE
        elif self.voice is not None:
            return enums.MessageType.VOICE
        elif self.contact is not None:
            return enums.MessageType.CONTACT
        elif self.dice is not None:
            return enums.MessageType.DICE
        elif self.game is not None:
            return enums.MessageType.GAME
        elif self.poll is not None:
            return enums.MessageType.POLL
        elif self.venue is not None:
            return enums.MessageType.VENUE
        elif self.location is not None:
            return enums.MessageType.LOCATION
        elif self.new_chat_members is not None:
            return enums.MessageType.NEW_CHAT_MEMBERS
        elif self.left_chat_member is not None:
            return enums.MessageType.LEFT_CHAT_MEMBER
        elif self.new_chat_title is not None:
            return enums.MessageType.NEW_CHAT_TITLE
        elif self.new_chat_photo is not None:
            return enums.MessageType.NEW_CHAT_PHOTO
        elif self.delete_chat_photo is not None:
            return enums.MessageType.DELETE_CHAT_PHOTO
        elif self.group_chat_created is not None:
            return enums.MessageType.GROUP_CHAT_CREATED
        elif self.supergroup_chat_created is not None:
            return enums.MessageType.SUPERGROUP_CHAT_CREATED
        elif self.channel_chat_created is not None:
            return enums.MessageType.CHANNEL_CHAT_CREATED
        elif self.message_auto_delete_timer_changed is not None:
            return enums.MessageType.MESSAGE_AUTO_DELETE_TIMER_CHANGED
        elif self.migrate_to_chat_id is not None:
            return enums.MessageType.MIGRATE_TO_CHAT_ID
        elif self.migrate_from_chat_id is not None:
            return enums.MessageType.MIGRATE_FROM_CHAT_ID
        elif self.pinned_message is not None:
            return enums.MessageType.PINNED_MESSAGE
        elif self.invoice is not None:
            return enums.MessageType.INVOICE
        elif self.successful_payment is not None:
            return enums.MessageType.SUCCESSFUL_PAYMENT
        elif self.connected_website is not None:
            return enums.MessageType.CONNECTED_WEBSITE
        elif self.passport_data is not None:
            return enums.MessageType.PASSPORT_DATA
        elif self.proximity_alert_triggered is not None:
            return enums.MessageType.PROXIMITY_ALERT_TRIGGERED
        elif self.video_chat_scheduled is not None:
            return enums.MessageType.VIDEO_CHAT_SCHEDULED
        elif self.video_chat_started is not None:
            return enums.MessageType.VIDEO_CHAT_STARTED
        elif self.video_chat_ended is not None:
            return enums.MessageType.VIDEO_CHAT_ENDED
        elif self.video_chat_participants_invited is not None:
            return enums.MessageType.VIDEO_CHAT_PARTICIPANTS_INVITED
        elif self.web_app_data is not None:
            return enums.MessageType.WEB_APP_DATA

        raise UnknownMessageTypeError("Unknown message type!")

    @property
    def is_forwarded(self) -> bool:
        return self.forward_date is not None

    @property
    def url(self) -> Optional[str]:
        if self.chat.type in {chat_types.CHANNEL, chat_types.SUPERGROUP}:
            if self.chat.username is not None:
                return get_message_public_url(self.chat.username, self.message_id)
            else:
                return get_message_private_url(self.chat.id, self.message_id)



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.chat import Chat
from telebox.telegram.types.types.user import User
from telebox.telegram.types.types.chat_invite_link import ChatInviteLink


@dataclass(unsafe_hash=True)
class ChatJoinRequest(Type):
    chat: Chat
    from_: User
    date: datetime
    bio: Optional[str] = None
    invite_link: Optional[ChatInviteLink] = None



from typing import Optional


def get_url(path: Optional[str] = None) -> str:
    url = f"https://t.me/"

    if path is not None:
        url += path

    return url


def get_user_mention_url(id_: int) -> str:
    return f"tg://user?id={id_}"


def get_message_public_url(path: str, message_id: int) -> str:
    return get_url(f"{path}/{message_id}")


def get_message_private_url(chat_id: int, message_id: int) -> str:
    return get_url(f"c/{chat_id}/{message_id}")


def get_chat_id_with_prefix(id_: int) -> int:
    if id_ > 0:
        id_without_prefix = id_
        multiplier = 1

        while id_without_prefix:
            id_without_prefix //= 10
            multiplier *= 10

        return -id_ + -100 * multiplier
    elif id_ == 0:
        return -100_0

    return id_


def get_chat_id_without_prefix(id_: int) -> int:
    if id_ <= -100_0:
        id_with_prefix = id_
        multiplier = 1

        while id_with_prefix != -100:
            id_with_prefix = -(id_with_prefix // -10)
            multiplier *= 10

            if not id_with_prefix:
                break
        else:
            return -100 * multiplier - id_

    return id_


def get_full_name(first_name: str, last_name: Optional[str]) -> str:
    full_name = first_name

    if last_name is not None:
        full_name += f" {last_name}"

    return full_name


def get_attach_string(name: str) -> str:
    return f"attach://{name}"



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type


@dataclass(unsafe_hash=True)
class Voice(Type):
    file_id: str
    file_unique_id: str
    duration: int
    mime_type: Optional[str] = None
    file_size: Optional[int] = None



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.message import Message
from telebox.telegram.types.types.inline_query import InlineQuery
from telebox.telegram.types.types.chosen_inline_result import ChosenInlineResult
from telebox.telegram.types.types.callback_query import CallbackQuery
from telebox.telegram.types.types.shipping_query import ShippingQuery
from telebox.telegram.types.types.pre_checkout_query import PreCheckoutQuery
from telebox.telegram.types.types.poll import Poll
from telebox.telegram.types.types.poll_answer import PollAnswer
from telebox.telegram.types.types.chat_member_updated import ChatMemberUpdated
from telebox.telegram.types.types.chat_join_request import ChatJoinRequest


@dataclass(unsafe_hash=True)
class Update(Type):
    update_id: int
    message: Optional[Message] = None
    edited_message: Optional[Message] = None
    channel_post: Optional[Message] = None
    edited_channel_post: Optional[Message] = None
    inline_query: Optional[InlineQuery] = None
    chosen_inline_result: Optional[ChosenInlineResult] = None
    callback_query: Optional[CallbackQuery] = None
    shipping_query: Optional[ShippingQuery] = None
    pre_checkout_query: Optional[PreCheckoutQuery] = None
    poll: Optional[Poll] = None
    poll_answer: Optional[PollAnswer] = None
    my_chat_member: Optional[ChatMemberUpdated] = None
    chat_member: Optional[ChatMemberUpdated] = None
    chat_join_request: Optional[ChatJoinRequest] = None



from dataclasses import dataclass
from typing import Literal, Optional

from telebox.telegram.types.base import Type


@dataclass(unsafe_hash=True)
class ForceReply(Type):
    force_reply: Literal[True] = True
    input_field_placeholder: Optional[str] = None
    selective: Optional[bool] = None



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type


@dataclass(unsafe_hash=True)
class WebhookInfo(Type):
    url: str
    has_custom_certificate: bool
    pending_update_count: int
    ip_address: Optional[str] = None
    last_error_date: Optional[int] = None
    last_error_message: Optional[str] = None
    last_synchronization_error_date: Optional[int] = None
    max_connections: Optional[int] = None
    allowed_updates: Optional[list[str]] = None



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.photo_size import PhotoSize


@dataclass(unsafe_hash=True)
class UserProfilePhotos(Type):
    total_count: int
    photos: list[list[PhotoSize]]



from typing import Union

from telebox.telegram.types.types.input_media_photo import InputMediaPhoto
from telebox.telegram.types.types.input_media_video import InputMediaVideo
from telebox.telegram.types.types.input_media_animation import InputMediaAnimation
from telebox.telegram.types.types.input_media_audio import InputMediaAudio
from telebox.telegram.types.types.input_media_document import InputMediaDocument


InputMedia = Union[InputMediaPhoto,
                   InputMediaVideo,
                   InputMediaAnimation,
                   InputMediaAudio,
                   InputMediaDocument]



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.consts import bot_command_scope_types


@dataclass(unsafe_hash=True)
class BotCommandScopeAllChatAdministrators(Type):
    type: str = bot_command_scope_types.ALL_CHAT_ADMINISTRATORS



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type


@dataclass(unsafe_hash=True)
class ChatPermissions(Type):
    can_send_messages: Optional[bool] = None
    can_send_media_messages: Optional[bool] = None
    can_send_polls: Optional[bool] = None
    can_send_other_messages: Optional[bool] = None
    can_add_web_page_previews: Optional[bool] = None
    can_change_info: Optional[bool] = None
    can_invite_users: Optional[bool] = None
    can_pin_messages: Optional[bool] = None



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.consts import passport_element_error_sources


@dataclass(unsafe_hash=True)
class PassportElementErrorDataField(Type):
    type: str
    field_name: str
    data_hash: str
    message: str
    source: str = passport_element_error_sources.DATA



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.encrypted_passport_element import EncryptedPassportElement
from telebox.telegram.types.types.encrypted_credentials import EncryptedCredentials


@dataclass(unsafe_hash=True)
class PassportData(Type):
    data: list[EncryptedPassportElement]
    credentials: EncryptedCredentials



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.keyboard_button_poll_type import KeyboardButtonPollType
from telebox.telegram.types.types.web_app_info import WebAppInfo


@dataclass(unsafe_hash=True)
class KeyboardButton(Type):
    text: str
    request_contact: Optional[bool] = None
    request_location: Optional[bool] = None
    request_poll: Optional[KeyboardButtonPollType] = None
    web_app: Optional[WebAppInfo] = None



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.user import User


@dataclass(unsafe_hash=True)
class PollAnswer(Type):
    poll_id: str
    user: User
    option_ids: list[int]



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.consts import input_media_types
from telebox.telegram.types.types.message_entity import MessageEntity


@dataclass(unsafe_hash=True)
class InputMediaPhoto(Type):
    media: str
    caption: Optional[str] = None
    parse_mode: Optional[str] = None
    caption_entities: Optional[list[MessageEntity]] = None
    type: str = input_media_types.PHOTO



from dataclasses import dataclass
from typing import Optional, Literal

from telebox.telegram.types.base import Type
from telebox.telegram.utils import get_url, get_user_mention_url, get_full_name


@dataclass(unsafe_hash=True)
class User(Type):
    id: int
    is_bot: bool
    first_name: str
    last_name: Optional[str] = None
    username: Optional[str] = None
    language_code: Optional[str] = None
    is_premium: Optional[Literal[True]] = None
    added_to_attachment_menu: Optional[Literal[True]] = None
    can_join_groups: Optional[bool] = None
    can_read_all_group_messages: Optional[bool] = None
    supports_inline_queries: Optional[bool] = None

    @property
    def full_name(self) -> str:
        return get_full_name(self.first_name, self.last_name)

    @property
    def url(self) -> Optional[str]:
        if self.username is not None:
            return get_url(self.username)

    @property
    def mention_url(self) -> str:
        return get_user_mention_url(self.id)



from dataclasses import dataclass
from typing import Union

from telebox.telegram.types.base import Type
from telebox.telegram.consts import bot_command_scope_types


@dataclass(unsafe_hash=True)
class BotCommandScopeChatAdministrators(Type):
    chat_id: Union[int, str]
    type: str = bot_command_scope_types.CHAT_ADMINISTRATORS



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.user import User
from telebox.telegram.types.types.message import Message


@dataclass(unsafe_hash=True)
class CallbackQuery(Type):
    id: str
    from_: User
    chat_instance: str
    message: Optional[Message] = None
    inline_message_id: Optional[str] = None
    data: Optional[str] = None
    game_short_name: Optional[str] = None



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.consts import inline_query_result_types
from telebox.telegram.types.types.inline_keyboard_markup import InlineKeyboardMarkup
from telebox.telegram.types.types.input_message_content import InputMessageContent


@dataclass(unsafe_hash=True)
class InlineQueryResultContact(Type):
    id: str
    phone_number: str
    first_name: str
    last_name: Optional[str] = None
    vcard: Optional[str] = None
    reply_markup: Optional[InlineKeyboardMarkup] = None
    input_message_content: Optional[InputMessageContent] = None
    thumb_url: Optional[str] = None
    thumb_width: Optional[int] = None
    thumb_height: Optional[int] = None
    type: str = inline_query_result_types.CONTACT



from dataclasses import dataclass
from typing import Union, Optional

from telebox.telegram.types.base import Type
from telebox.telegram.consts import input_media_types
from telebox.telegram.types.types.input_file import InputFile
from telebox.telegram.types.types.message_entity import MessageEntity


@dataclass(unsafe_hash=True)
class InputMediaAnimation(Type):
    media: str
    thumb: Union[InputFile, str, None] = None
    caption: Optional[str] = None
    parse_mode: Optional[str] = None
    caption_entities: Optional[list[MessageEntity]] = None
    width: Optional[int] = None
    height: Optional[int] = None
    duration: Optional[int] = None
    type: str = input_media_types.ANIMATION



from .telegram.telegram import Telegram


def _evaluate_type_annotations() -> None:
    from typing import get_type_hints

    from .telegram.types.types.chat import Chat
    from .telegram.types.types.message import Message

    classes = (Chat, Message)
    mapping = {i.__name__: i for i in classes}

    for class_ in classes:
        class_.__annotations__ = get_type_hints(class_, mapping)

        for name, type_ in class_.__annotations__.items():
            class_.__dataclass_fields__[name].type = type_  # noqa


def _patch_requests_serializer() -> None:
    import warnings

    import ujson

    # https://github.com/psf/requests/issues/1595#issuecomment-30993198
    try:
        import requests.compat
    except ModuleNotFoundError:
        warnings.warn("Failed to set ujson in requests: the structure has changed!", stacklevel=2)
    else:
        requests.compat.json = ujson


_evaluate_type_annotations()
_patch_requests_serializer()



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.user import User


@dataclass(unsafe_hash=True)
class ProximityAlertTriggered(Type):
    traveler: User
    watcher: User
    distance: int



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.consts import chat_member_statuses
from telebox.telegram.types.types.user import User


@dataclass(unsafe_hash=True)
class ChatMemberRestricted(Type):
    user: User
    is_member: bool
    can_change_info: bool
    can_invite_users: bool
    can_pin_messages: bool
    can_send_messages: bool
    can_send_media_messages: bool
    can_send_polls: bool
    can_send_other_messages: bool
    can_add_web_page_previews: bool
    until_date: Optional[datetime] = None  # None instead of 0
    status: str = chat_member_statuses.RESTRICTED



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.consts import menu_button_types


@dataclass(unsafe_hash=True)
class MenuButtonCommands(Type):
    type: str = menu_button_types.COMMANDS



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.consts import chat_member_statuses
from telebox.telegram.types.types.user import User


@dataclass(unsafe_hash=True)
class ChatMemberOwner(Type):
    user: User
    is_anonymous: bool
    custom_title: Optional[str] = None
    status: str = chat_member_statuses.CREATOR



from dataclasses import dataclass

from telebox.telegram.types.base import Type
from telebox.telegram.consts import chat_member_statuses
from telebox.telegram.types.types.user import User


@dataclass(unsafe_hash=True)
class ChatMemberMember(Type):
    user: User
    status: str = chat_member_statuses.MEMBER



PERSONAL_DETAILS = "personal_details"
PASSPORT = "passport"
DRIVER_LICENSE = "driver_license"
IDENTITY_CARD = "identity_card"
INTERNAL_PASSPORT = "internal_passport"
ADDRESS = "address"
UTILITY_BILL = "utility_bill"
BANK_STATEMENT = "bank_statement"
RENTAL_AGREEMENT = "rental_agreement"
PASSPORT_REGISTRATION = "passport_registration"
TEMPORARY_REGISTRATION = "temporary_registration"
PHONE_NUMBER = "phone_number"
EMAIL = "email"



from typing import Union

from telebox.telegram.types.types.chat_member_owner import ChatMemberOwner
from telebox.telegram.types.types.chat_member_administrator import ChatMemberAdministrator
from telebox.telegram.types.types.chat_member_member import ChatMemberMember
from telebox.telegram.types.types.chat_member_restricted import ChatMemberRestricted
from telebox.telegram.types.types.chat_member_left import ChatMemberLeft
from telebox.telegram.types.types.chat_member_banned import ChatMemberBanned


ChatMember = Union[ChatMemberOwner,
                   ChatMemberAdministrator,
                   ChatMemberMember,
                   ChatMemberRestricted,
                   ChatMemberLeft,
                   ChatMemberBanned]



from dataclasses import dataclass
from typing import Optional

from telebox.telegram.types.base import Type
from telebox.telegram.types.types.labeled_price import LabeledPrice


@dataclass(unsafe_hash=True)
class InputInvoiceMessageContent(Type):
    title: str
    description: str
    payload: str
    provider_token: str
    currency: str
    prices: list[LabeledPrice]
    max_tip_amount: Optional[int] = None
    suggested_tip_amounts: Optional[list[int]] = None
    provider_data: Optional[str] = None
    photo_url: Optional[str] = None
    photo_size: Optional[int] = None
    photo_width: Optional[int] = None
    photo_height: Optional[int] = None
    need_name: Optional[bool] = None
    need_phone_number: Optional[bool] = None
    need_email: Optional[bool] = None
    need_shipping_address: Optional[bool] = None
    send_phone_number_to_provider: Optional[bool] = None
    send_email_to_provider: Optional[bool] = None
    is_flexible: Optional[bool] = None
